// Copyright (c) 2006, TopCoder, Inc. All rights reserved.
using System;
using System.Collections.Generic;
using System.Xml;

namespace TopCoder.Uml.Xmi.MetaData
{
    /// <summary>
    /// Represents a method in a class.  Contains properties and a list of parameters.
    /// </summary>
    /// <author>Dan Manges (Seiz3r)</author>
    /// <copyright>Copyright (c) 2006, TopCoder, Inc. All rights reserved.</copyright>
    public class Method : XmiElement
    {
        /// <summary>
        /// Properties of the method.
        /// </summary>
        private string name, visibility, ownerScope;

        /// <summary>
        /// Boolean indicating if the method is abstract.
        /// </summary>
        private bool isAbstract;

        /// <summary>
        /// The return parameter for the method.  May be null indicating a constructor method.
        /// </summary>
        private UmlParameter returnParameter;

        /// <summary>
        /// The class this method belongs to.
        /// </summary>
        private UmlClass umlClass;

        /// <summary>
        /// List of parameter node id references, used until parsing XMI is complete.
        /// </summary>
        private List<string> parameterReferences = new List<string>();

        /// <summary>
        /// List of the method parameters.
        /// </summary>
        private List<UmlParameter> parameters = new List<UmlParameter>();

        /// <summary>
        /// Create a method from a UML:Operation node, reading XML attributes.
        /// </summary>
        /// <param name="node">The node to use to create the method object.</param>
        public Method(XmlNode node)
            : base(node)
        {
            // Read the name, visibility, scope, and determine if the method is abstract.
            this.name = Helper.ReadAttribute(node, "name");
            this.visibility = Helper.ReadAttribute(node, "visibility");
            this.ownerScope = Helper.ReadAttribute(node, "ownerScope");
            this.isAbstract = Helper.ReadAttribute(node, "isAbstract") == "true" ? true : false;
        }

        /// <summary>
        /// Add a parameter reference to this method.
        /// </summary>
        /// <param name="refId">The reference id.</param>
        public void AddParameterReference(string refId)
        {
            this.parameterReferences.Add(refId);
        }

        /// <summary>
        /// Get an array of parameter id references.
        /// </summary>
        public string[] ParameterReferences
        {
            get
            {
                return this.parameterReferences.ToArray();
            }
        }

        /// <summary>
        /// Get an array of method parameters.
        /// </summary>
        public UmlParameter[] Parameters
        {
            get
            {
                return this.parameters.ToArray();
            }
        }

        /// <summary>
        /// Add a paramter to this method.  Will either set the return parameter, or add to the list of parameters.
        /// </summary>
        /// <param name="p"></param>
        public void AddParameter(UmlParameter p)
        {
            if (p.Kind == "return")
            {
                this.returnParameter = p;
            }
            else
            {
                this.parameters.Add(p);
            }
        }

        /// <summary>
        /// Get the name of the method.
        /// </summary>
        /// <value>The name of the method.</value>
        public string Name
        {
            get
            {
                return this.name;
            }
        }

        /// <summary>
        /// Get or set the class the method belongs to.
        /// </summary>
        /// <value>The class the method belongs to.</value>
        /// <returns>The class the method belongs to.</returns>
        public UmlClass UmlClass
        {
            get
            {
                return this.umlClass;
            }
            set
            {
                this.umlClass = value;
            }
        }

        /// <summary>
        /// Set a string for the default implementation for the method.
        /// </summary>
        /// <returns></returns>
        private string GetImplementation()
        {
            if (this.umlClass == null)
            {
                return "";
            }
            if (this.umlClass.IsInterface || this.isAbstract)
            {
                // Interface and abstract methods have no implementation.
                return ";";
            }
            else
            {
                // Return implementation throwing a NotImplementedException.
                return "\r\n        {\r\n            throw new NotImplementedException();\r\n        }";
            }
        }

        /// <summary>
        /// Create a string of parameters as part of the method signature.  This simplifies generating the source code.
        /// </summary>
        /// <returns>A string of parameters for the method.</returns>
        private string ParametersString()
        {
            string signature = "";
            foreach (UmlParameter p in this.Parameters)
            {
                signature += String.Format("{0} {1}, ", p.Type, p.Name);
            }
            return signature.TrimEnd(new char[] { ',',' ' });
        }

        /// <summary>
        /// Gets any attribute modifiers read from stereotypes in the XMI file.
        /// </summary>
        /// <remarks>Currently supports readonly, virtual, and override.</remarks>
        /// <returns>Attribute modifier string.</returns>
        private string GetModifiers()
        {
            string[] possibleTypes = new string[] { "readonly", "virtual", "override" };
            foreach (string possibleType in possibleTypes)
            {
                if (Array.IndexOf<string>(this.Stereotypes, possibleType) > -1)
                {
                    return possibleType + " ";
                }
            }
            return " ";
        }

        /// <summary>
        /// Write the method data to an XmlWriter
        /// </summary>
        /// <param name="writer">The XmlWriter to write the data to.</param>
        public void WriteXml(XmlWriter writer)
        {
            writer.WriteStartElement("method");
            writer.WriteStartElement("documentation");
            string documentation = Helper.GetDocumentation(this, 12);
            foreach (UmlParameter p in this.Parameters)
            {
                documentation += Helper.GetDocumentation(p, 12);
            }
            if (this.returnParameter != null)
            {
                documentation += Helper.GetDocumentation(this.returnParameter, 12);
            }
            writer.WriteCData(documentation);
            writer.WriteEndElement();
            writer.WriteElementString("visibility", this.visibility);
            writer.WriteElementString("ownerScope", Helper.OwnerScopeToCSharp(this.ownerScope));
            writer.WriteElementString("abstract", this.isAbstract ? "abstract " : "");
            writer.WriteElementString("modifiers", this.GetModifiers());
            writer.WriteElementString("name", this.Name);
            writer.WriteElementString("returnType", this.returnParameter != null ? this.returnParameter.Type : "");
            writer.WriteElementString("parameters", this.ParametersString());
            writer.WriteStartElement("implementation");
            writer.WriteCData(this.GetImplementation());
            writer.WriteEndElement();
            writer.WriteStartElement("parameters");
            foreach (UmlParameter p in this.parameters)
            {
                p.WriteXml(writer);
            }
            writer.WriteEndElement();
            writer.WriteEndElement();
        }
    }
}